body {
  background-color: white;
  color: #343260;
  font-family: Jost, sans-serif;
  font-weight: 400;
  font-size: 1.2rem;
  line-height: 1.2
}
header {
  background: linear-gradient(to bottom, #ffffff 92%, #d9dfe3 92%, #ffffff 100%);
  padding-bottom: 5px;
}
a { color: #343260; text-decoration: underline; }
a:hover, a:active { color: #6280dc; }
a:visited { color: #536f9f }
h1 { font-size: 2rem; }
#form {
  padding: 2rem;
  background: linear-gradient(to bottom, #ffffff 0%, #d9dfe3 1%, #f3f3f6 1%, #f3f3f6 99%, #d9dfe3 99%, #ffffff 100%);
}
#text-box-wrapper {
  position: relative;
}
#text-box button {
  position: absolute;
  top: 10px;
  right: 10px;
  background-color: #ff9182;
  color: black;
  border: none;
  border-radius: 0px;
  padding: 2px 7px;
}
label, #suggestions {
  border-top: 1px solid #343260;
  padding-top: 0.5rem;
  text-transform: uppercase;
  font-size: 1.1rem;
  display: block;
}
.form-control { border-radius: 0px; }
select {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
}
.select-wrapper { position: relative; }
.select-wrapper:after {
  content: '▼';
  font-size: 10px;
  position: absolute;
  top: 11px;
  right: 9px;
  color: black;
  pointer-events: none;
}
#limit-buttons label {
  border-radius: 50%;
  margin-right: 0.4rem;
  width: 2rem;
  height: 2rem;
  display: inline-flex;
  align-items: center;
  justify-content: center;
  background-color: #c7d9ed;
  color: #343260;
  border-width: 0px;
  font-weight: 500;
}
#limit-buttons .active {
  background-color: #6280dc;
  color: white;
}
#get-suggestions {
  margin: 2rem 0;
  background-color: #6280dc;
  color: white;
  border: none;
  border-radius: 0px;
  padding-right: 3rem;
  background-image: url('/static/img/arrow-white.svg');
  background-position: 97% center;
  background-repeat: no-repeat;
}
#get-suggestions:disabled {
  cursor: default;
  background-color: #6c757d;
  color: white;
}
#results li, #no-results li { border-radius: 0; }
meter { width: 24px; }
meter:-moz-meter-optimum::-moz-meter-bar {
  background: #6280dc;
}
meter::-webkit-meter-bar {
  border: none;
  border-radius: 0;
  height: 18px;
  background-color: #ccc;
  box-shadow: 0 12px 3px -5px #e6e6e6 inset;
}
meter::-webkit-meter-optimum-value {
  background: #6280dc;
}
